@@ -1,4 +1,4 @@
-#  Copyright (c) 2009, Cloud Matrix Pty. Ltd.
+#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
 #  All rights reserved; available under the terms of the BSD License.
 """
 
@@ -9,8 +9,8 @@ a simple API through which apps can find, fetch and install updates, and a
 bootstrapping mechanism that keeps the app safe in the face of failed or
 partial updates.
 
-Esky is currently capable of freezing apps with bbfreeze, cxfreeze, py2exe and
-py2app. Adding support for other freezer programs should be straightforward;
+Esky is currently capable of freezing apps with py2exe, py2app, cxfreeze and
+bbfreeze. Adding support for other freezer programs should be straightforward;
 patches will be gratefully accepted.
 
 The main interface is the 'Esky' class, which represents a frozen app.  An Esky
@@ -21,16 +21,16 @@ for an app automatically updating itself would look something like this:
     if hasattr(sys,"frozen"):
         app = esky.Esky(sys.executable,"http://example.com/downloads/")
         app.auto_update()
-        app.cleanup()
 
 A simple default VersionFinder is provided that hits a specified URL to get
 a list of available versions.  More sophisticated implementations will likely
 be added in the future, and you're encouraged to develop a custom VersionFinder
 subclass to meet your specific needs.
 
-To freeze your application in a format suitable for use with esky, use the
-"bdist_esky" distutils command.  See the docstring of esky.bdist_esky for
-full details; the following is an example of a simple setup.py using esky:
+The real trick is freezing your app in a format sutiable for use with esky.
+You'll almost certainly want to use the "bdist_esky" distutils command, and
+should consult its docstring for full details; the following is an example
+of a simple setup.py script using esky:
 
     from esky import bdist_esky
     from distutils.core import setup
@@ -54,7 +54,45 @@ The contents of this zipfile can be extracted to the filesystem to give a
 fully working application.  If made available online then it can also be found,
 downloaded and used as an upgrade by older versions of the application.
 
-The on-disk layout of an app managed by esky looks like this:
+
+When you find you need to move beyond the simple logic of Esky.auto_update()
+(e.g. to show feedback in the GUI) then the following properties and methods
+and available on the Esky class:
+
+    app.version:                the current best available version.
+
+    app.active_version:         the currently-executing version, or None
+                                if the esky isn't for the current app.
+
+    app.find_update():          find the best available update, or None
+                                if no updates are available.
+
+    app.fetch_version(v):       fetch the specified version into local storage.
+
+    app.install_version(v):     install and activate the specified version.
+
+    app.uninstall_version(v):   (try to) uninstall the specified version; will
+                                fail if the version is currently in use.
+
+    app.cleanup():              (try to) clean up various partly-installed
+                                or old versions lying around the app dir.
+
+    app.reinitialize():         re-initialize internal state after changing
+                                the installed version.
+
+If updating an application that is not writable by normal users, esky has the
+ability to gain root privileges through the use of a helper program.  The
+following methods control this behaviour:
+
+    app.has_root():             check whether esky currently has root privs.
+
+    app.get_root():             escalate to root privs by spawning helper app.
+
+    app.drop_root():            kill helper app and drop root privileges
+
+
+When properly installed, the on-disk layout of an app managed by esky looks
+like this:
 
     prog.exe                 - esky bootstrapping executable
     updates/                 - work area for fetching/unpacking updates
@@ -62,21 +100,25 @@ The on-disk layout of an app managed by esky looks like this:
         prog.exe             - executable(s) as produced by freezer module
         library.zip          - pure-python frozen modules
         pythonXY.dll         - python DLL
+        esky-bootstrap/      - files not yet moved into bootstrapping env
         esky-bootstrap.txt   - list of files expected in the bootstrapping env
+        esky-lockfile.txt    - lock file to control access to in-use versions
         ...other deps...
 
 This is also the layout of the zipfiles produced by bdist_esky.  The 
 "appname-X.Y" directory is simply a frozen app directory with some extra
-bootstrapping information in the file "esky-bootstrap.txt".
+control information generated by esky.
 
-To upgrade to a new version "appname-X.Z", esky performs the following steps:
+To install a new version "appname-X.Z", esky performs the following steps:
     * extract it into a temporary directory under "updates"
     * move all bootstrapping files into "appname-X.Z.platm/esky-bootstrap"
     * atomically rename it into the main directory as "appname-X.Z.platform"
     * move contents of "appname-X.Z.platform/esky-bootstrap" into the main dir
     * remove the "appname-X.Z.platform/esky-bootstrap" directory
-    * remove files not in "appname-X.Z.platform/esky-bootstrap.txt"
-    * remove the "appname-X.Y.platform" directory
+
+To uninstall an existing version "appname-X.Y", esky does the following
+    * remove files used by only that version from the bootstrap env
+    * rename its "esky-bootstrap.txt" file to "esky-bootstrap-old.txt"
 
 Where such facilities are provided by the operating system, this process is
 performed within a filesystem transaction. Nevertheless, the esky bootstrapping
@@ -84,14 +126,15 @@ executable is able to detect and recover from a failed update should such an
 unfortunate situation arise.
 
 To clean up after failed or partial updates, applications should periodically
-call the "cleanup" method on their esky.
+call the "cleanup" method on their esky.  This removes uninstalled versions
+and generally tries to tidy up in the main application directory.
 
 """
 
 from __future__ import with_statement
 
 __ver_major__ = 0
-__ver_minor__ = 6
+__ver_minor__ = 7
 __ver_patch__ = 0
 __ver_sub__ = ""
 __version__ = "%d.%d.%d%s" % (__ver_major__,__ver_minor__,__ver_patch__,__ver_sub__)
@@ -103,7 +146,15 @@ import shutil
 import errno
 import socket
 import time
+import subprocess
+import atexit
 from functools import wraps
+from base64 import b64encode, b64decode
+
+try:
+    import cPickle as pickle
+except ImportError:
+    import pickle
 
 try:
     import threading
@@ -112,12 +163,13 @@ except ImportError:
 
 if sys.platform != "win32":
     import fcntl
+else:
+    from esky.winres import is_safe_to_overwrite
 
 from esky.errors import *
-from esky.fstransact import FSTransaction
+from esky.fstransact import FSTransaction, files_differ
 from esky.finder import DefaultVersionFinder
-import esky.helper
-from esky.helper import EskyHelperApp
+from esky.sudo import SudoProxy, has_root, allow_from_sudo
 from esky.util import split_app_version, join_app_version,\
                       is_version_dir, is_uninstalled_version_dir,\
                       parse_version, get_best_version, appdir_from_executable,\
@@ -125,20 +177,6 @@ from esky.util import split_app_version, join_app_version,\
 
 
 
-def use_helper_app(func):
-    """Method decorator to transparently use an esky's helper app, if present.
-
-    This decorator wraps an Esky method so that it is transparently proxied
-    to the helper process whenever the esky's "helper_app" attribute is set.
-    """
-    @wraps(func)
-    def method_using_helper_app(self,*args,**kwds):
-        if self.helper_app is not None:
-            return getattr(self.helper_app,func.func_name)(*args,**kwds)
-        return func(self,*args,**kwds)
-    return method_using_helper_app
-
-
 
 class Esky(object):
     """Class representing an updatable frozen app.
@@ -150,7 +188,6 @@ class Esky(object):
         if hasattr(sys,"frozen"):
             app = esky.Esky(sys.executable,"http://example.com/downloads/")
             app.auto_update()
-            app.cleanup()
 
     The first argument must be either the top-level application directory,
     or the path of an executable from that application.  The second argument
@@ -162,6 +199,21 @@ class Esky(object):
     lock_timeout = 60*60  # 1 hour
 
     def __init__(self,appdir_or_exe,version_finder=None):
+        self._init_from_appdir(appdir_or_exe)
+        self._lock_count = 0
+        self.sudo_proxy = None
+        self.keep_sudo_proxy_alive = False
+        self._old_sudo_proxies = []
+        self.version_finder = version_finder
+        self.reinitialize()
+
+    def _init_from_appdir(self,appdir_or_exe):
+        """Extension point to override the initial logic of Esky initialisation.
+
+        This method is expected to interrogate the given appdir and set up the
+        basic properties of the esky (e.g. name, platform) in response.  It is
+        split into its own method to make it easier to override in subclasses.
+        """
         if os.path.isfile(appdir_or_exe):
             self.appdir = appdir_from_executable(appdir_or_exe)
             vdir = appdir_or_exe[len(self.appdir):].split(os.sep)[1]
@@ -170,11 +222,6 @@ class Esky(object):
         else:
             self.active_version = None
             self.appdir = appdir_or_exe
-        self.reinitialize()
-        self._lock_count = 0
-        self.version_finder = version_finder
-        self.helper_app = None
-        self.HelperAppClass = EskyHelperApp
 
     def _get_version_finder(self):
         return self.__version_finder
@@ -197,7 +244,7 @@ class Esky(object):
             v = join_app_version(self.name,self.active_version,self.platform)
         else:
             v = join_app_version(self.name,self.version,self.platform)
-        return os.path.abspath(oss.path.join(self.appdir,v,relpath))
+        return os.path.abspath(os.path.join(self.appdir,v,relpath))
 
     def reinitialize(self):
         """Reinitialize internal state by poking around in the app directory.
@@ -222,6 +269,9 @@ class Esky(object):
         Locking is achieved by creating a "locked" directory and writing the
         current process/thread ID into it.  os.mkdir is atomic on all platforms
         that we care about. 
+
+        This also has the side-effect of failing early if the user does not
+        have permission to modify the application directory.
         """
         if num_retries > 5:
             raise EskyLockedError
@@ -286,29 +336,48 @@ class Esky(object):
             os.unlink(os.path.join(lockdir,myid))
             os.rmdir(lockdir)
 
-    @use_helper_app
+    @allow_from_sudo()
     def has_root(self):
         """Check whether the user currently has root/administrator access."""
-        return esky.helper.has_root()
+        if self.sudo_proxy is not None:
+            return self.sudo_proxy.has_root()
+        return has_root()
 
     def get_root(self):
         """Attempt to gain root/administrator access by spawning helper app."""
         if self.has_root():
             return True
-        self.helper_app = self.HelperAppClass(self,as_root=True)
-        if not self.helper_app.has_root():
+        self.sudo_proxy = SudoProxy(self)
+        self.sudo_proxy.start()
+        if not self.sudo_proxy.has_root():
             raise OSError(None,"could not escalate to root privileges")
 
-    @use_helper_app
+    def drop_root(self):
+        """Drop root privileges by killing the helper app."""
+        if self.sudo_proxy is not None:
+            self.sudo_proxy.close()
+            if self.keep_sudo_proxy_alive:
+                self._old_sudo_proxies.append(self.sudo_proxy)
+            else:
+                self.sudo_proxy.terminate()
+            self.sudo_proxy = None
+
+    @allow_from_sudo()
     def cleanup(self):
         """Perform cleanup tasks in the app directory.
 
         This includes removing older versions of the app and completing any
         failed update attempts.  Such maintenance is not done automatically
         since it can take a non-negligible amount of time.
+
+        If the cleanup proceeds sucessfully this method will return True; it
+        there is work that cannot currently be completed, it returns False.
         """
+        if self.sudo_proxy is not None:
+            return self.sudo_proxy.cleanup()
         appdir = self.appdir
         self.lock()
+        success = True
         try:
             best_version = get_best_version(appdir)
             new_version = get_best_version(appdir,include_partial_installs=True)
@@ -326,7 +395,8 @@ class Esky(object):
             manifest.add("updates")
             manifest.add("locked")
             manifest.add(best_version)
-            if self.active_version:
+            if self.active_version and self.active_version != best_version:
+                success = False
                 manifest.add(self.active_version)
             for nm in os.listdir(appdir):
                 if nm not in manifest:
@@ -338,23 +408,78 @@ class Esky(object):
                         try:
                             self.uninstall_version(v)
                         except VersionLockedError:
-                            pass
+                            success = False
                         else:
-                            self._try_remove(appdir,nm,manifest)
+                            success &= self._try_remove(appdir,nm,manifest)  
                     elif is_uninstalled_version_dir(fullnm):
                         #  It's a partially-removed version; finish removing it.
-                        self._try_remove(appdir,nm,manifest)
+                        success &= not self._try_remove(appdir,nm,manifest)
                     elif ".old." in nm or nm.endswith(".old"):
                         #  It's a temporary backup file; remove it.
-                        self._try_remove(appdir,nm,manifest)
+                        success &= not self._try_remove(appdir,nm,manifest)
                     else:
                         #  It's an unaccounted-for entry in the bootstrap env.
                         #  Can't prove it's safe to remove, so leave it.
                         pass
+            #  If there are pending overwrites, try to do them.
+            ovrdir = os.path.join(appdir,best_version,"esky-overwrite")
+            if os.path.exists(ovrdir):
+                try:
+                    for (dirnm,_,filenms) in os.walk(ovrdir,topdown=False):
+                        for nm in filenms:
+                            ovrsrc = os.path.join(dirnm,nm)
+                            ovrdst = os.path.join(appdir,ovrsrc[len(ovrdir)+1:])
+                            with open(ovrsrc,"rb") as fIn:
+                                with open(ovrdst,"ab") as fOut:
+                                    fOut.seek(0)
+                                    chunk = fIn.read(512*16)
+                                    while chunk:
+                                        fOut.write(chunk)
+                                        chunk = fIn.read(512*16)
+                            os.unlink(ovrsrc)
+                        os.rmdir(dirnm)
+                except EnvironmentError, e:
+                    success = False
+            #  Get the VersionFinder to clean up after itself
             if self.version_finder is not None:
                 self.version_finder.cleanup(self)
         finally:
             self.unlock()
+        return success
+
+    @allow_from_sudo()
+    def cleanup_at_exit(self):
+        """Arrange for cleanup to occur after application exit.
+
+        This operates by using the atexit module to spawn a new instance of 
+        this app, with appropriate flags that cause it to launch directly into
+        the cleanup process.
+ 
+        Recall that sys.executable points to a specific version dir, so this
+        new process will not hold any filesystem locks in the main app dir.
+        """
+        if self.sudo_proxy is not None:
+            self.keep_sudo_proxy_alive = True
+            return self.sudo_proxy.cleanup_at_exit()
+        if not getattr(sys,"frozen",False):
+            exe = [sys.executable,"-c","import esky; esky.run_startup_hooks()","--esky-spawn-cleanup"]
+        elif os.path.basename(sys.executable).lower() in ("python","pythonw"):
+            exe = [sys.executable,"-c","import esky; esky.run_startup_hooks()","--esky-spawn-cleanup"]
+        else:
+            if not _startup_hooks_were_run:
+                raise OSError(None,"unable to sudo: startup hooks not run")
+            exe = [sys.executable,"--esky-spawn-cleanup"]
+        exe = exe + [b64encode(pickle.dumps(self,pickle.HIGHEST_PROTOCOL))]
+        @atexit.register
+        def spawn_cleanup():
+            rnul = open(os.devnull,"r")
+            wnul = open(os.devnull,"w")
+            if sys.platform == "win32":
+                kwds = dict(close_fds=True)
+            else:
+                kwds = dict(stdin=rnul,stdout=wnul,stderr=wnul,close_fds=True)
+            subprocess.Popen(exe,**kwds)
+        
 
     def _try_remove(self,appdir,path,manifest=[]):
         """Try to remove the file/directory at the given path in the appdir.
@@ -370,19 +495,23 @@ class Esky(object):
         """
         fullpath = os.path.join(appdir,path)
         if fullpath in sys.path:
-            return
+            return False
         if path in manifest:
-            return
+            return False
         try:
             if os.path.isdir(fullpath):
                 #  Remove paths starting with "esky-" last, since we use
                 #  these to maintain state information.
                 esky_paths = []
+                success = True
                 for nm in os.listdir(fullpath):
                     if nm.startswith("esky-"):
                         esky_paths.append(nm)
                     else:
-                        self._try_remove(appdir,os.path.join(path,nm),manifest)
+                        subdir = os.path.join(path,nm)
+                        success &= self._try_remove(appdir,subdir,manifest)
+                if not success:
+                    return False
                 for nm in sorted(esky_paths):
                     self._try_remove(appdir,os.path.join(path,nm),manifest)
                 os.rmdir(fullpath)
@@ -391,42 +520,87 @@ class Esky(object):
         except EnvironmentError, e:
             if e.errno not in self._errors_to_ignore:
                 raise
+            return False
+        else:
+            return True
     _errors_to_ignore = (errno.ENOENT, errno.EPERM, errno.EACCES, errno.ENOTDIR,
                          errno.EISDIR, errno.EINVAL, errno.ENOTEMPTY,)
 
     def auto_update(self):
-        """Automatically install the latest version of the app."""
+        """Automatically install the latest version of the app.
+
+        This method automatically performs the following sequence of actions,
+        escalating to root privileges if a permission error is encountered:
+
+            * find the latest version [self.find_update()]
+            * fetch the new version [self.fetch_version()]
+            * install the new version [self.install_version()]
+            * attempt to uninstall the old version [self.uninstall_version()]
+            * reinitialize internal state [self.reinitialize()]
+            * clean up the appdir [self.cleanup()]
+
+        This method is mostly here to help you get started.  For an app of
+        any serious complexity, you will probably want to build your own
+        variant that e.g. operates in a background thread, prompts the user
+        for confirmation, etc.
+        """
         if self.version_finder is None:
             raise NoVersionFinderError
-        version = self.find_update()
-        if version is not None:
-            assert parse_version(version) > parse_version(self.version)
-            #  Try to install the new version.  If it fails with
-            #  a permission error, escalate to root and try again.
-            try:
-                self.fetch_version(version)
-                self.install_version(version)
+        got_root = False
+        cleaned = False
+        try:
+            version = self.find_update()
+            if version is not None:
+                #  Try to install the new version.  If it fails with
+                #  a permission error, escalate to root and try again.
                 try:
-                    self.uninstall_version(self.version)
-                except VersionLockedError:
-                    pass
-            except EnvironmentError, e:
-                if e.errno != errno.EACCES or self.has_root():
-                    raise
+                    self._do_auto_update(version)
+                except EnvironmentError:
+                    exc_type,exc_value,exc_traceback = sys.exc_info()
+                    if exc_value.errno != errno.EACCES or self.has_root():
+                        raise
+                    try:
+                        self.get_root()
+                    except Exception, e:
+                        raise exc_type,exc_value,exc_traceback
+                    else:
+                        got_root = True
+                        self._do_auto_update(version)
+                self.reinitialize()
+            #  Try to clean up the app dir.  If it fails with a 
+            #  permission error, escalate to root and try again.
+            try:
+                cleaned = self.cleanup()
+            except EnvironmentError:
                 exc_type,exc_value,exc_traceback = sys.exc_info()
+                if exc_value.errno != errno.EACCES or self.has_root():
+                    raise
                 try:
                     self.get_root()
                 except Exception, e:
                     raise exc_type,exc_value,exc_traceback
                 else:
-                    self.fetch_version(version)
-                    self.install_version(version)
-                    try:
-                        self.uninstall_version(self.version)
-                    except VersionLockedError:
-                        pass
-            else:
-                self.reinitialize()
+                    got_root = True
+                    cleaned = self.cleanup()
+        finally:
+            #  Drop root privileges as soon as possible.
+            if not cleaned:
+                self.cleanup_at_exit()
+            if got_root:
+                self.drop_root()
+
+    def _do_auto_update(self,version):
+        """Actual sequence of operations for auto-update.
+
+        This is a separate method so it can easily be retried after gaining
+        root privileges.
+        """
+        self.fetch_version(version)
+        self.install_version(version)
+        try:
+            self.uninstall_version(self.version)
+        except VersionLockedError:
+            pass
 
     def find_update(self):
         """Check for an available update to this app.
@@ -445,11 +619,17 @@ class Esky(object):
                 best_version = version
         return best_version
 
-    @use_helper_app
+    @allow_from_sudo(str)
     def fetch_version(self,version):
         """Fetch the specified updated version of the app."""
+        if self.sudo_proxy is not None:
+            return self.sudo_proxy.fetch_version(version)
         if self.version_finder is None:
             raise NoVersionFinderError
+        #  Guard against malicious input (might be called with root privs)
+        target = join_app_version(self.name,version,self.platform)
+        target = os.path.join(self.appdir,target)
+        assert os.path.dirname(target) == self.appdir
         #  Get the new version using the VersionFinder
         loc = self.version_finder.has_version(self,version)
         if not loc:
@@ -459,7 +639,7 @@ class Esky(object):
         copy_ownership_info(os.path.join(self.appdir,vdir),loc)
         return loc
 
-    @use_helper_app
+    @allow_from_sudo(str)
     def install_version(self,version):
         """Install the specified version of the app.
 
@@ -467,9 +647,13 @@ class Esky(object):
         available as a version directory inside the app directory.  It 
         does not modify any other installed versions.
         """
+        if self.sudo_proxy is not None:
+            return self.sudo_proxy.install_version(version)
         #  Extract update then rename into position in main app directory
         target = join_app_version(self.name,version,self.platform)
         target = os.path.join(self.appdir,target)
+        #  Guard against malicious input (might be called with root privs)
+        assert os.path.dirname(target) == self.appdir
         if not os.path.exists(target):
             self.fetch_version(version)
             source = self.version_finder.has_version(self,version)
@@ -479,18 +663,7 @@ class Esky(object):
                 os.rename(source,target)
             trn = FSTransaction()
             try:
-                #  Move new bootrapping environment into main app dir.
-                #  Be sure to move dependencies before executables.
-                bootstrap = os.path.join(target,"esky-bootstrap")
-                with open(os.path.join(target,"esky-bootstrap.txt"),"rt") as f:
-                    for nm in f:
-                        nm = nm.strip()
-                        bssrc = os.path.join(bootstrap,nm)
-                        bsdst = os.path.join(self.appdir,nm)
-                        if os.path.exists(bssrc):
-                            trn.move(bssrc,bsdst)
-                #  Remove the bootstrap dir; the new version is now installed
-                trn.remove(bootstrap)
+                self._unpack_bootstrap_env(version,trn)
             except Exception:
                 trn.abort()
                 raise
@@ -499,11 +672,49 @@ class Esky(object):
         finally:
             self.unlock()
 
-    @use_helper_app
+    def _unpack_bootstrap_env(self,version,trn):
+        """Unpack the bootstrap env from the given target directory."""
+        vdir = join_app_version(self.name,version,self.platform)
+        target = os.path.join(self.appdir,vdir)
+        assert os.path.dirname(target) == self.appdir
+        #  Move new bootrapping environment into main app dir.
+        #  Be sure to move dependencies before executables.
+        bootstrap = os.path.join(target,"esky-bootstrap")
+        for nm in self._version_manifest(vdir):
+            bssrc = os.path.join(bootstrap,nm)
+            bsdst = os.path.join(self.appdir,nm)
+            if os.path.exists(bssrc):
+                #  On windows we can't atomically replace files.
+                #  If they differ in a "safe" way we put them aside
+                #  to overwrite at a later time.
+                if sys.platform == "win32" and os.path.exists(bsdst):
+                    if not files_differ(bssrc,bsdst):
+                        trn.remove(bssrc)
+                    elif is_safe_to_overwrite(bssrc,bsdst):
+                        ovrdir = os.path.join(target,"esky-overwrite")
+                        if not os.path.exists(ovrdir):
+                            os.mkdir(ovrdir)
+                        trn.move(bssrc,os.path.join(ovrdir,nm))
+                    else:
+                        trn.move(bssrc,bsdst)
+                else:
+                    trn.move(bssrc,bsdst)
+            if os.path.isdir(os.path.dirname(bssrc)):
+                if not os.listdir(os.path.dirname(bssrc)):
+                    trn.remove(os.path.dirname(bssrc))
+        #  Remove the bootstrap dir; the new version is now installed
+        trn.remove(bootstrap)
+
+    @allow_from_sudo(str)
     def uninstall_version(self,version): 
         """Uninstall the specified version of the app."""
+        if self.sudo_proxy is not None:
+            return self.sudo_proxy.uninstall_version(version)
         target_name = join_app_version(self.name,version,self.platform)
         target = os.path.join(self.appdir,target_name)
+        #  Guard against malicious input (might be called with root privs)
+        assert os.path.dirname(target) == self.appdir
+        lockfile = os.path.join(target,"esky-lockfile.txt")
         bsfile = os.path.join(target,"esky-bootstrap.txt")
         bsfile_old = os.path.join(target,"esky-bootstrap-old.txt")
         self.lock()
@@ -515,36 +726,17 @@ class Esky(object):
             try:
                 trn = FSTransaction()
                 try:
-                    #  Get set of all files that must stay in the main appdir
-                    to_keep = set()
-                    for vname in os.listdir(self.appdir):
-                        if vname == target_name:
-                            continue
-                        details = split_app_version(vname)
-                        if details[0] != self.name:
-                            continue
-                        if parse_version(details[1]) < parse_version(version):
-                            continue
-                        to_keep.update(self._version_manifest(vname))
-                    #  Remove files used only by the version being removed
-                    to_rem = self._version_manifest(target_name) - to_keep
-                    for nm in to_rem:
-                        fullnm = os.path.join(self.appdir,nm)
-                        if os.path.exists(fullnm):
-                            trn.remove(fullnm)
+                    self._cleanup_bootstrap_env(version,trn)
                 except Exception:
                     trn.abort()
                     raise
                 else:
                     trn.commit()
             except EnvironmentError:
-                try:
-                    open(bsfile,"a").close()
-                except EnvironmentError:
+                if is_locked_version_dir(target):
                     raise VersionLockedError("version in use: %s" % (version,))
-                else:
-                    raise
-            #  Disable the version by removing its esky-bootstrap.txt file.
+                raise
+            #  Disable the version by renaming its esky-bootstrap.txt file.
             #  To avoid clobbering in-use version, respect locks on this file.
             if sys.platform == "win32":
                 try:
@@ -552,7 +744,7 @@ class Esky(object):
                 except EnvironmentError:
                     raise VersionLockedError("version in use: %s" % (version,))
             else:
-                f = open(bsfile,"r")
+                f = open(lockfile,"r")
                 try:
                     fcntl.flock(f,fcntl.LOCK_EX|fcntl.LOCK_NB)
                 except EnvironmentError, e:
@@ -568,60 +760,63 @@ class Esky(object):
         finally:
             self.unlock()
 
+    def _cleanup_bootstrap_env(self,version,trn):
+        """Cleanup the bootstrap env populated by the given version."""
+        target_name = join_app_version(self.name,version,self.platform)
+        #  Get set of all files that must stay in the main appdir
+        to_keep = set()
+        for vname in os.listdir(self.appdir):
+            if vname == target_name:
+                continue
+            details = split_app_version(vname)
+            if details[0] != self.name:
+                continue
+            if parse_version(details[1]) < parse_version(version):
+                continue
+            to_keep.update(self._version_manifest(vname))
+        #  Remove files used only by the version being removed
+        to_rem = self._version_manifest(target_name) - to_keep
+        for nm in to_rem:
+            fullnm = os.path.join(self.appdir,nm)
+            if os.path.exists(fullnm):
+                trn.remove(fullnm)
+            if os.path.isdir(os.path.dirname(fullnm)):
+                if not os.listdir(os.path.dirname(fullnm)):
+                    trn.remove(os.path.dirname(fullnm))
+
     def _version_manifest(self,vdir):
         """Get the bootstrap manifest for the given version directory.
 
         This is the set of files/directories that the given version expects
-        to be in the main app directory
+        to be in the main app directory.
         """
         mpath = os.path.join(self.appdir,vdir,"esky-bootstrap.txt")
+        manifest = set()
         try:
             with open(mpath,"rt") as mf:
-                return set(ln.strip() for ln in mf)
+                for ln in mf:
+                    #  Guard against malicious input, since we might try
+                    #  to manipulate these files with root privs.
+                    nm = os.path.normpath(ln.strip())
+                    assert not os.path.isabs(nm)
+                    assert not nm.startswith("..")
+                    manifest.add(nm)
         except IOError:
-            return set()
+            pass
+        return manifest
 
 
 
-def _check_needsroot(func):
-    @wraps(func)
-    def do_check_needsroot(self,*args,**kwds):
-        if os.environ.get("ESKY_NEEDSROOT",""):
-            if not self.has_root():
-                raise OSError(errno.EACCES,"you need root")
-        return func(self,*args,**kwds)
-    return do_check_needsroot
-
-
-class _TestableEsky(Esky):
-    """Esky subclass that tries harder to be testable.
-
-    If the environment variable "ESKY_NEEDSROOT" is set, operations that
-    alter the filesystem will fail with EACCES when not executed as root.
-    """
-
-    @_check_needsroot
-    def lock(self,num_retries=0):
-        return super(_TestableEsky,self).lock(num_retries)
-
-    @_check_needsroot
-    def unlock(self):
-        return super(_TestableEsky,self).unlock()
-   
-    @_check_needsroot
-    def cleanup(self):
-        return super(_TestableEsky,self).cleanup()
-
-    @_check_needsroot
-    def fetch_version(self,version):
-        return super(_TestableEsky,self).fetch_version(version)
-
-    @_check_needsroot
-    def install_version(self,version):
-        return super(_TestableEsky,self).install_version(version)
-
-    @_check_needsroot
-    def uninstall_version(self,version):
-        super(_TestableEsky,self).uninstall_version(version)
+_startup_hooks_were_run = False
 
+def run_startup_hooks():
+    global _startup_hooks_were_run
+    _startup_hooks_were_run = True
+    if len(sys.argv) > 1 and sys.argv[1] == "--esky-spawn-cleanup":
+        app = pickle.loads(b64decode(sys.argv[2]))
+        time.sleep(10)        
+        app.cleanup()
+        sys.exit(0)
+    import esky.sudo
+    esky.sudo.run_startup_hooks()
 
