@@ -1,4 +1,4 @@
-#  Copyright (c) 2009, Cloud Matrix Pty. Ltd.
+#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
 #  All rights reserved; available under the terms of the BSD License.
 """
 
@@ -46,14 +46,7 @@ def freeze(dist):
     tdir = tempfile.mkdtemp()
     try:
         for exe in dist.get_executables():
-            #  bbfreeze doesn't offer a way to set the name of the target exe.
-            #  We write each to a tempfile to force the selected name.
-            name = exe.name
-            if sys.platform == "win32" and name.endswith(".exe"):
-                name = name[:-4]
-            script = os.path.join(tdir,name+".py")
-            shutil.copy2(exe.script,script)
-            f.addScript(script,gui_only=exe.gui_only)
+            f.addScript(exe.script,gui_only=exe.gui_only)
         if "include_py" not in options:
             f.include_py = False
         if "linkmethod" not in options:
@@ -119,7 +112,7 @@ def freeze(dist):
     #  Copy the loader program for each script.
     #  We explicitly strip the loader binaries, in case they were made
     #  by linking to the library.zip.
-    for exe in dist.get_executables():
+    for exe in dist.get_executables(rewrite=False):
         if not exe.include_in_bootstrap_env:
             continue
         exepath = dist.copy_to_bootstrap_env(exe.name)
@@ -151,7 +144,15 @@ def _chainload(target_dir):
       else:
           sys.modules.pop("esky",None)
           sys.modules.pop("esky.bootstrap",None)
-          exec code in {"__name__":"__main__"}
+          try:
+              exec code in {"__name__":"__main__"}
+          except zipimport.ZipImportError, e:
+              #  If it can't find the __main__{sys.executable} script,
+              #  the user might be running from a backup exe file.
+              #  Fall back to original chainloader to attempt workaround.
+              if e.message.startswith("can't find module '__main__"):
+                  _orig_chainload(target_dir)
+              raise
           sys.exit(0)
 """
 
