@@ -1,4 +1,4 @@
-#  Copyright (c) 2009, Cloud Matrix Pty. Ltd.
+#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
 #  All rights reserved; available under the terms of the BSD License.
 """
 
@@ -22,7 +22,9 @@ use during the bootstrap process:
   Chainloading:         execv, chainload
   Filesystem:           listdir, exists, basename, dirname, pathjoin
   Version handling:     split_app_version, join_app_version, parse_version,
-                        get_all_versions, get_best_version
+                        get_all_versions, get_best_version, is_version_dir,
+                        is_installed_version_dir, is_uninstalled_version_dir,
+                        lock_version_dir, unlock_version_dir
 
 
 """
@@ -40,6 +42,11 @@ elif "nt" in sys.builtin_module_names:
     fcntl = None
     from nt import listdir, stat, unlink, rename, spawnv, P_WAIT
     SEP = "\\"
+    #  The standard execv terminates the spawning process, which makes
+    #  it impossible to wait for it.  This alternative is waitable, but
+    #  risks leaving zombie children if it is killed externally.
+    #  TODO: some way to kill children when this is killed - should be doable
+    #        with some custom code in child startup.
     def execv(filename,args):
         res = spawnv(P_WAIT,filename,args)
         raise SystemExit(res)
@@ -115,26 +122,12 @@ def chainload(target_dir):
     it will not be removed by any simultaneously-running instances of the
     application.
     """
-    #  This global will hold the locked file to keep it open during execution.
-    global _version_dir_lockfile
-    lockfile = pathjoin(target_dir,"esky-bootstrap.txt")
     try:
-        #  On windows, holding the file open is enough to lock it.
-        #  On other platforms, try for a shared lock using fcntl.flock.
-        #  While fcntl.fcntl locks are apparently the new hotness, they have
-        #  unfortunate semantics that we don't want for this application:
-        #      * LOCK_EX requires write access to the file
-        #      * not inherited across fork()
-        #      * released when closing *any* fd associated with that file
-        #  fcntl.flock doesn't have these problems, but may fail on NFS.
-        #  I'll wait for the bug reports to come in on that one...
-        _version_dir_lockfile = open(lockfile,"r")
-        if fcntl is not None:
-            fcntl.flock(_version_dir_lockfile,fcntl.LOCK_SH)
+        lock_version_dir(target_dir)
     except EnvironmentError:
-        #  If the lockfile has gone missing, the version is being uninstalled.
+        #  If the bootstrap file is missing, the version is being uninstalled.
         #  Our only option is to re-execute ourself and find the new version.
-        if exists(lockfile):
+        if exists(pathjoin(target_dir,"esky-bootstrap.txt")):
             raise
         execv(sys.executable,sys.argv)
     else:
@@ -150,7 +143,16 @@ def _chainload(target_dir):
     """
     appdir = dirname(target_dir)
     target_exe = target_dir + sys.executable[len(appdir):]
-    execv(target_exe,[target_exe] + sys.argv[1:])
+    try:
+        execv(target_exe,[target_exe] + sys.argv[1:])
+    except EnvironmentError, e:
+        if e.errno == errno.ENOENT:
+            # Tried to chainload something that doesn't exist.
+            # Perhaps executing from a backup file?
+            orig_exe = get_original_filename(sys.executable)
+            if orig_exe is not None:
+                target_exe = target+dir + orig_executable[len(appdir):]
+                execv(target_exe,[target_exe] + sys.argv[1:])
 
 
 def get_best_version(appdir,include_partial_installs=False):
@@ -330,3 +332,49 @@ def _split_version_components(s):
         yield s[start:end]
         start = end
 
+
+def get_original_filename(backname):
+    """Given a backup filename, get the original name to which it refers.
+
+    This is only really possible if the original file actually exists and
+    is not guaranteed to be correct in all cases; but unless you do something
+    silly it should work out OK.
+
+    If no matching original file is found, None is returned.
+    """
+    filtered = ".".join(filter(lambda n: n != "old",backname.split(".")))
+    for nm in listdir(dirname(backname)):
+        if nm == backname:
+            continue
+        if filtered == ".".join(filter(lambda n: n != "old",nm.split("."))):
+            return pathjoin(dirname(backname),nm)
+    return None
+
+
+_locked_version_dirs = {}
+
+def lock_version_dir(vdir):
+    """Lock the given version dir so it cannot be uninstalled."""
+    if sys.platform == "win32":
+        #  On win32, we just hold bootstrap file open for reading.
+        #  This will prevent it from being renamed during uninstall.
+        lockfile = pathjoin(vdir,"esky-bootstrap.txt")
+        _locked_version_dirs.setdefault(vdir,[]).append(open(lockfile,"rt"))
+    else:
+        #  On posix platforms we take a shared flock on esky-lockfile.txt.
+        #  While fcntl.fcntl locks are apparently the new hotness, they have
+        #  unfortunate semantics that we don't want for this application:
+        #      * not inherited across fork()
+        #      * released when closing *any* fd associated with that file
+        #  fcntl.flock doesn't have these problems, but may fail on NFS.
+        #  To complicate matters, python sometimes emulated flock with fcntl!
+        #  We therefore use a separate lock file to avoid unpleasantness.
+        lockfile = pathjoin(vdir,"esky-lockfile.txt")
+        f = open(lockfile,"r")
+        _locked_version_dirs.setdefault(vdir,[]).append(f)
+        fcntl.flock(f,fcntl.LOCK_SH)
+
+def unlock_version_dir(vdir):
+    """Unlock the given version dir, allowing it to be uninstalled."""
+    _locked_version_dirs[vdir].pop().close()
+
